@@ -1,16 +1,25 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 
+/*
+ * By design, this file can be read without reading config.h
+ * #include "config.h" must appear as the first line of your .cpp file.
+ */
+
+#ifndef PACKAGE_NAME
+#error bulk_extractor_i.h included before config.h
+#endif
+
 #ifndef BULK_EXTRACTOR_I_H
 #define BULK_EXTRACTOR_I_H
 
-#define DEBUG_PEDANTIC    0x0001	// check values more rigorously
+#define DEBUG_PEDANTIC    0x0001        // check values more rigorously
 #define DEBUG_PRINT_STEPS 0x0002        // prints as each scanner is started
-#define DEBUG_SCANNER     0x0004	// dump all feature writes to stderr
-#define DEBUG_NO_SCANNERS 0x0008        /* do not run the scanners */
-#define DEBUG_DUMP_DATA   0x0010	/* dump data as it is seen */
-#define DEBUG_INFO        0x0040	// print extra info
-#define DEBUG_EXIT_EARLY  1000		/* just print the size of the volume and exis */
-#define DEBUG_ALLOCATE_512MiB 1002	/* Allocate 512MiB, but don't set any flags */
+#define DEBUG_SCANNER     0x0004        // dump all feature writes to stderr
+#define DEBUG_NO_SCANNERS 0x0008        // do not run the scanners
+#define DEBUG_DUMP_DATA   0x0010        // dump data as it is seen
+#define DEBUG_INFO        0x0040        // print extra info
+#define DEBUG_EXIT_EARLY  1000          // just print the size of the volume and exis 
+#define DEBUG_ALLOCATE_512MiB 1002      // Allocate 512MiB, but don't set any flags 
 
 /* We need netinet/in.h or windowsx.h */
 #ifdef HAVE_NETINET_IN_H
@@ -54,7 +63,6 @@
  * The scanners respond by setting fields in the SP and returning.
  * 
  * When executing, once again each scanner is called with the SP and RCB.
- * By design, this file can be read without reading config.h
  * This is the only file that needs to be included for a scanner.
  *
  * \li \c phase_startup - scanners are loaded and register the names of the feature files they want.
@@ -84,7 +92,6 @@ namespace be13 {
 #include "feature_recorder.h"
 #include "feature_recorder_set.h"
 
-
 /* Network includes */
 
 /****************************************************************
@@ -135,45 +142,6 @@ namespace be13 {
  * @param fs   - where the features should be saved. Must be provided if feature_names==0.
  **/
 
-class histogram_def {
- public:
-    /**
-     * @param feature- the feature file to histogram (no .txt)
-     * @param re     - the regular expression to extract
-     * @param require- require this string on the line (usually in context)
-     * @param suffix - the suffix to add to the histogram file after feature name before .txt
-     * @param flags  - any flags (see above)
-     */
-
-    histogram_def(string feature_,string re_,string suffix_,uint32_t flags_=0):
-        feature(feature_),pattern(re_),require(),suffix(suffix_),flags(flags_){}
-    histogram_def(string feature_,string re_,string require_,string suffix_,uint32_t flags_=0):
-        feature(feature_),pattern(re_),require(require_),suffix(suffix_),flags(flags_){}
-    string feature;                     /* feature file */
-    string pattern;                     /* extract pattern; "" means use entire feature */
-    string require;
-    string suffix;                      /* suffix to append; "" means "histogram" */
-    uint32_t flags;                     // defined in histogram.h
-};
-
-typedef  set<histogram_def> histograms_t;
-
-inline bool operator <(class histogram_def h1,class histogram_def h2)  {
-    if (h1.feature<h2.feature) return true;
-    if (h1.feature>h2.feature) return false;
-    if (h1.pattern<h2.pattern) return true;
-    if (h1.pattern>h2.pattern) return false;
-    if (h1.suffix<h2.suffix) return true;
-    if (h1.suffix>h2.suffix) return false;
-    return false;                       /* equal */
-};
-
-inline bool operator !=(class histogram_def h1,class histogram_def h2)  {
-    return h1.feature!=h2.feature || h1.pattern!=h2.pattern || h1.suffix!=h2.suffix;
-};
-
-
-
 /*****************************************************************
  *** bulk_extractor has a private implementation of IPv4 and IPv6,
  *** UDP and TCP. 
@@ -408,17 +376,48 @@ public:
     
 #ifndef ETHERTYPE_PUP
 #define ETHERTYPE_PUP           0x0200          /* Xerox PUP */
+#endif
+
+#ifndef ETHERTYPE_SPRITE
 #define ETHERTYPE_SPRITE        0x0500          /* Sprite */
+#endif
+
+#ifndef ETHERTYPE_IP
 #define ETHERTYPE_IP            0x0800          /* IP */
+#endif
+
+#ifndef ETHERTYPE_ARP
 #define ETHERTYPE_ARP           0x0806          /* Address resolution */
+#endif
+
+#ifndef ETHERTYPE_REVARP
 #define ETHERTYPE_REVARP        0x8035          /* Reverse ARP */
+#endif
+
+#ifndef ETHERTYPE_AT
 #define ETHERTYPE_AT            0x809B          /* AppleTalk protocol */
+#endif
+
+#ifndef ETHERTYPE_AARP
 #define ETHERTYPE_AARP          0x80F3          /* AppleTalk ARP */
+#endif
+
+#ifndef ETHERTYPE_VLAN
 #define ETHERTYPE_VLAN          0x8100          /* IEEE 802.1Q VLAN tagging */
+#endif
+
+#ifndef ETHERTYPE_IPX
 #define ETHERTYPE_IPX           0x8137          /* IPX */
+#endif
+
+#ifndef ETHERTYPE_IPV6
 #define ETHERTYPE_IPV6          0x86dd          /* IP protocol version 6 */
+#endif
+
+#ifndef ETHERTYPE_LOOPBACK
 #define ETHERTYPE_LOOPBACK      0x9000          /* used to test interfaces */
 #endif
+
     
     inline u_short packet_info::nshort(const u_char *buf,size_t pos) 
     {
@@ -608,18 +607,11 @@ typedef void packet_callback_t(void *user,const be13::packet_info &pi);
 class scanner_info {
 private:
     static std::stringstream helpstream; // where scanner info help messages are saved.
-    class not_impl: public exception {
-        virtual const char *what() const throw() {
-            return "copying scanner_info objects is not implemented.";
-        }
-    };
-    scanner_info(const scanner_info &i) __attribute__((__noreturn__))
-    :si_version(),name(),author(),description(),url(),
-        scanner_version(),flags(),feature_names(),histogram_defs(),
-        packet_user(),packet_cb(),config(){
-        throw new not_impl();}
-    ;
-    const scanner_info &operator=(const scanner_info &i){ throw new not_impl();}
+
+    // default copy construction and assignment are meaningless
+    // and not implemented
+    scanner_info(const scanner_info &i);
+    scanner_info &operator=(const scanner_info &i);
  public:
     static std::string helpstr(){return helpstream.str();}
     typedef std::map<std::string,std::string>  config_t; // configuration for scanner passed in
@@ -634,7 +626,7 @@ private:
     static const int SCANNER_WANTS_NGRAMS   = 0x040; // v3: Scanner gets buffers that are constant n-grams
     static const int SCANNER_FAST_FIND      = 0x080; // v3: This scanner is a very fast FIND scanner
     static const int SCANNER_DEPTH_0        = 0x100; // v3: scanner only runs at detph 0 by default
-    static const int CURRENT_SI_VERSION=3;          // 
+    static const int CURRENT_SI_VERSION     = 4;     
 
     static const std::string flag_to_string(const int flag){
         std::string ret;
@@ -649,13 +641,16 @@ private:
         return ret;
     }
 
-    /* Global config is passed to each scanner as a pointer */
+    /* Global config is passed to each scanner as a pointer when it is loaded.
+     * Scanner histograms are added to 'histograms' by machinery.
+     */
     struct scanner_config {
-        scanner_config():namevals(),debug(),hasher(){};
+        scanner_config():namevals(),debug(),hasher(),histograms(){};
         virtual ~scanner_config(){}
-        config_t  namevals; // v3: (input) name=val map
-        int       debug; // v3: (input) current debug level
-        struct be13::hash_def  hasher;
+        config_t  namevals;             // v3: (input) name=val map
+        int       debug;                // v3: (input) current debug level
+        struct be13::hash_def  hasher;  // v3: (input) hasher to use
+        histograms_t    histograms;     // v4: output
     };
 
     // never change the order or delete old fields, or else you will
@@ -747,21 +742,19 @@ class scanner_params {
     scanner_params(phase_t phase_,const sbuf_t &sbuf_,class feature_recorder_set &fs_,
                    PrintOptions &print_options_):
         sp_version(CURRENT_SP_VERSION),
-        phase(phase_),sbuf(sbuf_),fs(fs_),depth(0),print_options(print_options_),info(0),sbufxml(0){
+        phase(phase_),sbuf(sbuf_),fs(fs_),depth(0),print_options(print_options_),info(0),sxml(0){
     }
 
     /* A scanner params with no print options */
-    scanner_params(phase_t phase_,const sbuf_t &sbuf_,
-                   class feature_recorder_set &fs_):
+    scanner_params(phase_t phase_,const sbuf_t &sbuf_, class feature_recorder_set &fs_):
         sp_version(CURRENT_SP_VERSION),
-        phase(phase_),sbuf(sbuf_),fs(fs_),depth(0),print_options(no_options),info(0),sbufxml(0){
+        phase(phase_),sbuf(sbuf_),fs(fs_),depth(0),print_options(no_options),info(0),sxml(0){
     }
 
-    /* A scanner params with no print options but xmlstream */
-    scanner_params(phase_t phase_,const sbuf_t &sbuf_,
-                   class feature_recorder_set &fs_,std::stringstream *xmladd):
+    /* A scanner params with no print options but an xmlstream */
+    scanner_params(phase_t phase_,const sbuf_t &sbuf_,class feature_recorder_set &fs_,std::stringstream *xmladd):
         sp_version(CURRENT_SP_VERSION),
-        phase(phase_),sbuf(sbuf_),fs(fs_),depth(0),print_options(no_options),info(0),sbufxml(xmladd){
+        phase(phase_),sbuf(sbuf_),fs(fs_),depth(0),print_options(no_options),info(0),sxml(xmladd){
     }
 
     /** Construct a scanner_params for recursion from an existing sp and a new sbuf.
@@ -770,7 +763,7 @@ class scanner_params {
     scanner_params(const scanner_params &sp_existing,const sbuf_t &sbuf_new):
         sp_version(CURRENT_SP_VERSION),phase(sp_existing.phase),
         sbuf(sbuf_new),fs(sp_existing.fs),depth(sp_existing.depth+1),
-        print_options(sp_existing.print_options),info(sp_existing.info),sbufxml(0){
+        print_options(sp_existing.print_options),info(sp_existing.info),sxml(0){
         assert(sp_existing.sp_version==CURRENT_SP_VERSION);
     };
 
@@ -786,11 +779,11 @@ class scanner_params {
     const phase_t phase;                 /* v1: 0=startup, 1=normal, 2=shutdown (changed to phase_t in v1.3) */
     const sbuf_t &sbuf;                 /* v1: what to scan / only valid in SCAN_PHASE */
     class feature_recorder_set &fs;     /* v1: where to put the results / only valid in SCAN_PHASE */
-    const uint32_t depth;               /* v1: how far down are we? / only valid in SCAN_PHASE */
+    const uint32_t depth;            /* v1: how far down are we? / only valid in SCAN_PHASE */
 
-    PrintOptions  &print_options;       /* v1: how to print / NOT USED IN SCANNERS */
-    scanner_info  *info;                /* v2: set/get parameters on startup */
-    std::stringstream *sbufxml;         /* v3: tags added to the sbuf's XML stream (advanced feature) */
+    PrintOptions  &print_options;    /* v1: how to print / NOT USED IN SCANNERS */
+    scanner_info  *info;             /* v2: set/get parameters on startup */
+    std::stringstream *sxml;         /* v3: on scanning and shutdown: CDATA added to XML stream (advanced feature) */
 };
 
 
@@ -834,15 +827,18 @@ namespace be13 {
 
         static void set_scanner_debug(int debug);
 
-        static void load_scanner(scanner_t scanner,const scanner_info::scanner_config &sc); // load a specific scanner
-        static void load_scanner_file(string fn,const scanner_info::scanner_config &sc);    // load a scanner from a file
-        static void load_scanners(scanner_t * const *scanners_builtin,const scanner_info::scanner_config &sc);           // load the scan_ plugins
-        static void load_scanner_directory(const string &dirname,const scanner_info::scanner_config &sc); // load scanners in the directory
-        static void load_scanner_directories(const std::vector<std::string> &dirnames,const scanner_info::scanner_config &sc);
+        static void load_scanner(scanner_t scanner,scanner_info::scanner_config &sc); // load a specific scanner
+        static void load_scanner_file(string fn,scanner_info::scanner_config &sc);    // load a scanner from a file
+        static void load_scanners(scanner_t * const *scanners_builtin,scanner_info::scanner_config &sc);           // load the scan_ plugins
+        static void load_scanner_directory(const string &dirname,scanner_info::scanner_config &sc); // load scanners in the directory
+        static void load_scanner_directories(const std::vector<std::string> &dirnames,scanner_info::scanner_config &sc);
         static void load_scanner_packet_handlers();
         
-        static void message_enabled_scanners(scanner_params::phase_t phase,feature_recorder_set *fs); // send every enabled scanner the phase message
-        static scanner_t *find_scanner(const std::string &name); // returns the named scanner, or 0 if no scanner of that name
+        // send every enabled scanner the phase message
+        static void message_enabled_scanners(scanner_params::phase_t phase,feature_recorder_set &fs);
+
+        // returns the named scanner, or 0 if no scanner of that name
+        static scanner_t *find_scanner(const std::string &name); 
         static void get_enabled_scanners(std::vector<std::string> &svector); // put the enabled scanners into the vector
         static bool find_scanner_enabled(); // return true if a find scanner is enabled
         
@@ -854,7 +850,7 @@ namespace be13 {
         static void scanners_enable(const std::string &name); // saves a command to enable this scanner
         static void scanners_disable(const std::string &name); // saves a command to disable this scanner
         static void scanners_process_enable_disable_commands();               // process the enable/disable and config commands
-        static void scanners_init(feature_recorder_set *fs); // init the scanners
+        static void scanners_init(feature_recorder_set &fs); // init the scanners
 
         static void info_scanners(bool detailed_info,
                                   bool detailed_settings,
@@ -862,12 +858,10 @@ namespace be13 {
         
 
         /* Run the phases on the scanners */
-        static void phase_shutdown(feature_recorder_set &fs);
-        typedef void (*xml_notifier_t)(const std::string &xmlstring);
-        static void phase_histogram(feature_recorder_set &fs, xml_notifier_t xml_error_notifier);
+        static void phase_shutdown(feature_recorder_set &fs,std::stringstream *sxml=0); // sxml is where to put XML from scanners that shutdown
         static uint32_t get_max_depth_seen();
         static void process_sbuf(const class scanner_params &sp);                              /* process for feature extraction */
-        static void process_packet_info(const be13::packet_info &pi);
+        static void process_packet(const be13::packet_info &pi);
 
         /* recorders */
         static void get_scanner_feature_file_names(feature_file_names_t &feature_file_names);
@@ -891,6 +885,11 @@ inline std::string safe_utf16to8(std::wstring s){ // needs to be cleaned up
     return utf8_line;
 }
 
+// truncate string at the matching char
+inline void truncate_at(std::string &line, char ch) {
+    size_t pos = line.find(ch);
+    if(pos != std::string::npos) line.resize(pos);
+}
 
 #ifndef HAVE_ISXDIGIT
 inline int isxdigit(int c)
@@ -929,8 +928,8 @@ inline std::string microsoftDateToISODate(const uint64_t &time)
 inline bool validASCIIName(const std::string &name)
 {
     for(size_t i = 0; i< name.size(); i++){
-	if(((u_char)name[i]) & 0x80) return false; // high bit should not be set
-	if(((u_char)name[i]) < ' ') return false;  // should not be control character
+        if(((u_char)name[i]) & 0x80) return false; // high bit should not be set
+        if(((u_char)name[i]) < ' ') return false;  // should not be control character
     }
     return true;
 }
