@@ -1,19 +1,10 @@
 /* -*- mode: C++; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 #include "config.h"
-
-#ifdef BULK_EXTRACTOR
-#include "bulk_extractor.h"
-#else
 #include "bulk_extractor_i.h"
-#endif
-
 #include "unicode_escape.h"
 #include "beregex.h"
-
-#ifdef USE_HISTOGRAMS
 #include "histogram.h"
-#endif
 
 #include <unistd.h>
 #include <fcntl.h>
@@ -35,7 +26,9 @@
 #define DEBUG_PEDANTIC    0x0001// check values more rigorously
 #endif
 
+#ifndef WIN32
 pthread_t feature_recorder::main_threadid = 0;
+#endif
 size_t  feature_recorder::context_window_default=16;                    /* number of bytes of context */
 int64_t feature_recorder::offset_add   = 0;
 std::string  feature_recorder::banner_file;
@@ -44,15 +37,6 @@ uint32_t feature_recorder::opt_max_feature_size=1024*1024;
 uint32_t feature_recorder::debug=0;
 
 
-/** 
- * 1. Put the UTF-8 BOM in the file.
- * 2. Read the contents of the file pointed to by opt_banner and put each line into "f"
- * 3. Add the header
- */
-
-//const string feature_recorder::UTF8_BOM("\xef\xbb\xbf");
-//const string feature_recorder::BOM_EXPLAINATION("# UTF-8 Byte Order Marker; see http://unicode.org/faq/utf_bom.html\n");
-
 void feature_recorder::banner_stamp(std::ostream &os,const std::string &header)
 {
     int banner_lines = 0;
@@ -95,21 +79,27 @@ void feature_recorder::banner_stamp(std::ostream &os,const std::string &header)
  * and thus a different feature recorder, to avoid locking
  * problems. 
  *
- * @param name - the name of the feature being recorded.
- * @param histogram_enabled - whether or not a histogram should be created.
+ * @param outdir_      - where the feature file is written
+ * @param input_fname_ - the file (disk image) that these features were extracted from.
+ *                     - We should probably have a callback function to annotate the feature file.
+ * @param name         - the name of the feature being recorded.
  */
 
-feature_recorder::feature_recorder(string outdir_,string input_fname_,string name_):
-    flags(0),histogram_enabled(false),
-    outdir(outdir_),input_fname(input_fname_),name(name_),ignore_encoding(),count_(0),ios(),
-    context_window_before(context_window_default),context_window_after(context_window_default),
-    Mf(),Mr(), 
+feature_recorder::feature_recorder(class feature_recorder_set &fs_,
+                                   const std::string &outdir_,const std::string &input_fname_,const std::string &name_):
+    flags(0),
+    outdir(outdir_),input_fname(input_fname_),name(name_),ignore_encoding(),ios(),
+    fs(fs_),
+    count_(0),context_window_before(context_window_default),context_window_after(context_window_default),
+    Mf(),Mr(),mhistogram(),
     stop_list_recorder(0),
-    file_number(0),carve_mode(CARVE_ENCODED)
+    file_number_(0),carve_mode(CARVE_ENCODED)
 {
 }
 
-/* Don't have to erase the stop_list_recorder because it is in the set */
+/* Don't have to delete the stop_list_recorder because it is in the
+ * feature_recorder_set and will be separately deleted.
+ */
 feature_recorder::~feature_recorder()
 {
     if(ios.is_open()){
@@ -155,7 +145,7 @@ void feature_recorder::open()
     // Just open the stream for output
     ios.open(fname.c_str(),ios_base::out);
     if(!ios.is_open()){
-        cerr << "*** CANNOT OPEN FEATURE FILE "
+        cerr << "*** feature_recorder::open CANNOT OPEN FEATURE FILE FOR WRITING "
              << fname << ":" << strerror(errno) << "\n";
         exit(1);
     }
@@ -183,7 +173,7 @@ static inline bool isodigit(char c)
 /* statics */
 const std::string feature_recorder::feature_file_header("# Feature-File-Version: 1.1\n");
 const std::string feature_recorder::histogram_file_header("# Histogram-File-Version: 1.1\n");
-const std::string feature_recorder::bulk_extractor_version_header("# "PACKAGE_NAME "-Version: " PACKAGE_VERSION " ($Rev: 10844 $)\n");
+const std::string feature_recorder::bulk_extractor_version_header("# " PACKAGE_NAME "-Version: " PACKAGE_VERSION " ($Rev: 10844 $)\n");
 
 static inline int hexval(char ch)
 {
@@ -236,19 +226,34 @@ string feature_recorder::unquote_string(const string &s)
     return line.substr(feature_start);  // no context to remove
 }
 
-#ifdef USE_HISTOGRAMS
+void feature_recorder::set_flag(uint32_t flags_)
+{
+    MAINTHREAD();
+    flags|=flags_;
+
+    if((flags & FLAG_MEM_HISTOGRAM) && mhistogram==0){
+        /* Create the in-memory histogram */
+        mhistogram = new mhistogram_t();
+    }
+}
+
+
 /**
  *  Create a histogram for this feature recorder and an extraction pattern.
  */
 
-static void truncate_at(string &line,char ch)
+void dump_cb(const std::string &str,const uint64_t &count)
 {
-    size_t pos = line.find(ch);
-    if(pos!=string::npos) line.erase(pos);
+    std::cerr << "dump_cb: " << str << " - " << count << "\n";
 }
 
-void feature_recorder::make_histogram(const class histogram_def &def)
+void feature_recorder::make_histogram(const class histogram_def &def,feature_recorder::callback_t cb)
 {
+    if(flag_set(FLAG_MEM_HISTOGRAM)){
+        mhistogram->dump_sorted(dump_cb);
+        return;
+    }
+
     beregex reg(def.pattern,REG_EXTENDED);
     string ifname = fname_counter("");  // source of features
 
@@ -333,7 +338,6 @@ void feature_recorder::make_histogram(const class histogram_def &def)
     std::cerr << "Looped " << max_histogram_files
               << " times on histogram; something seems wrong\n";
 }
-#endif
 
 
 /****************************************************************
@@ -385,6 +389,19 @@ void feature_recorder::printf(const char *fmt, ...)
 }
 
 
+/**
+ * Combine the pos0, feature and context into a single line and write it to the feature file.
+ */
+
+void feature_recorder::write0(const pos0_t &pos0,const string &feature,const string &context)
+{
+    stringstream ss;
+    ss << pos0.shift(feature_recorder::offset_add).str() << '\t' << feature;
+    if(flag_notset(FLAG_NO_CONTEXT) && (context.size()>0)) ss << '\t' << context;
+    this->write(ss.str());
+}
+
+
 /**
  * the main entry point of writing a feature and its context to the feature file.
  * processes the stop list
@@ -421,12 +438,12 @@ void feature_recorder::write(const pos0_t &pos0,const string &feature_,const str
     string feature = validateOrEscapeUTF8(feature_, escape_bad_utf8,escape_backslash);
 
     string context;
-    if((flags & FLAG_NO_CONTEXT)==0){
+    if(flag_notset(FLAG_NO_CONTEXT)){
         context = validateOrEscapeUTF8(context_,escape_bad_utf8,escape_backslash);
     }
 
-    if(feature.size() > opt_max_feature_size) feature = feature.substr(0,opt_max_feature_size);
-    if(context.size() > opt_max_context_size) context = context.substr(0,opt_max_context_size);
+    if(feature.size() > opt_max_feature_size) feature.resize(opt_max_feature_size);
+    if(context.size() > opt_max_context_size) context.resize(opt_max_context_size);
     if(feature.size()==0){
         cerr << "zero length feature at " << pos0 << "\n";
         if(debug & DEBUG_PEDANTIC) assert(0);
@@ -446,26 +463,24 @@ void feature_recorder::write(const pos0_t &pos0,const string &feature_,const str
         }
     }
         
-
-
-#ifdef USE_STOP_LIST
     /* First check to see if the feature is on the stop list.
      * Only do this if we have a stop_list_recorder (the stop list recorder itself
      * does not have a stop list recorder. If it did we would infinitely recurse.
      */
-    if(((flags & FLAG_NO_STOPLIST)==0) && stop_list_recorder){          
-        if(stop_list.check_feature_context(feature,context)){
+    if(flag_notset(FLAG_NO_STOPLIST) && stop_list_recorder){          
+        if(fs.stop_list
+           && fs.stop_list->check_feature_context(feature,context)){
             stop_list_recorder->write(pos0,feature,context);
             return;
         }
     }
-#endif
 
-#ifdef USE_ALERT_LIST
     /* The alert list is a special features that are called out.
      * If we have one of those, write it to the redlist.
      */
-    if(((flags & FLAG_NO_ALERTLIST)==0) && alert_list.check_feature_context(feature,context)){
+    if(flag_notset(FLAG_NO_ALERTLIST)
+       && fs.alert_list
+       && fs.alert_list->check_feature_context(feature,context)){
         string alert_fn = outdir + "/ALERTS_found.txt";
 
         cppmutex::lock lock(Mr);                // notce we are locking the redlist
@@ -474,13 +489,16 @@ void feature_recorder::write(const pos0_t &pos0,const string &feature_,const str
             rf << pos0.shift(feature_recorder::offset_add).str() << '\t' << feature << '\t' << "\n";
         }
     }
-#endif
+
+    /* Support in-memory histogram */
+    if(flag_set(FLAG_MEM_HISTOGRAM)){
+        mhistogram->add(feature,1);
+    }
 
     /* Finally write out the feature and the context */
-    stringstream ss;
-    ss << pos0.shift(feature_recorder::offset_add).str() << '\t' << feature;
-    if(((flags & FLAG_NO_CONTEXT)==0) && (context.size()>0)) ss << '\t' << context;
-    this->write(ss.str());
+    if(flag_notset(FLAG_NO_FEATURES)){
+        this->write0(pos0,feature,context);
+    }
 }
 
 /**
@@ -558,9 +576,30 @@ std::string replace(const std::string &src,char f,char t)
  ****************************************************************
  *
  * Carving support.
+ * 2013-08-29 - replace invalid characters in filenames
  * 2013-07-30 - automatically bin directories
  * 2013-06-08 - filenames are the forensic path.
  */
+
+std::string valid_dosname(std::string in)
+{
+    std::string out;
+    for(size_t i=0;i<in.size();i++){
+        uint8_t ch = in.at(i);
+        if(ch<=32 || ch>=128
+           || ch=='"' || ch=='*' || ch=='+' || ch==','
+           || ch=='/' || ch==':' || ch==';' || ch=='<'
+           || ch=='=' || ch=='>' || ch=='?' || ch=='\\'
+           || ch=='[' || ch==']' || ch=='|'){
+            out.push_back('_');
+        } else {
+            out.push_back(ch);
+        }
+    }
+    return out;
+}
+        
+
 #include <iomanip>
 /**
  * @param sbuf   - the buffer to carve
@@ -573,6 +612,8 @@ std::string feature_recorder::carve(const sbuf_t &sbuf,size_t pos,size_t len,
                                     const std::string &ext,
                                     const be13::hash_def &hasher)
 {
+    if(flags & FLAG_DISABLED) return std::string();           // disabled
+
     /* If we are in the margin, ignore; it will be processed again */
     if(pos >= sbuf.pagesize && pos < sbuf.bufsize){
         return std::string();
@@ -615,15 +656,14 @@ std::string feature_recorder::carve(const sbuf_t &sbuf,size_t pos,size_t len,
      * that's okay, because the second one will fail.
      */
 
-    uint64_t this_file_number = __sync_add_and_fetch(&file_number,1);
-
+    uint64_t this_file_number = file_number_add(1);
     std::string dirname1 = outdir + "/" + name;
     std::stringstream ss;
 
     ss << dirname1 << "/" << std::setw(3) << std::setfill('0') << (this_file_number / 1000);
 
     std::string dirname2 = ss.str(); 
-    std::string fname    = dirname2 + "/" + replace(sbuf.pos0.str(),'/','_') + ext;
+    std::string fname    = dirname2 + std::string("/") + valid_dosname(sbuf.pos0.str() + ext);
     std::string carved_hash_hexvalue = (*hasher.func)(sbuf.buf,sbuf.bufsize);
 
     /* Record what was found in the feature file.
@@ -669,8 +709,13 @@ std::string feature_recorder::carve(const sbuf_t &sbuf,size_t pos,size_t len,
     return fname;
 }
 
+/**
+ * Currently, we need strptime() and utimes() to set the time.
+ */
 void feature_recorder::set_carve_mtime(const std::string &fname, const std::string &mtime_iso8601) 
 {
+    if(flags & FLAG_DISABLED) return;           // disabled
+#if defined(HAVE_STRPTIME) && defined(HAVE_UTIMES)
     if(fname.size()){
         struct tm tm;
         if(strptime(mtime_iso8601.c_str(),"%Y-%m-%dT%H:%M:%S",&tm)){
@@ -681,33 +726,6 @@ void feature_recorder::set_carve_mtime(const std::string &fname, const std::stri
             }
         }
     }
+#endif
 }
 
-/**
- * Tagging support.
- * Tags are a system of providing identifiers for regions of the input image.
- * Currently tags are just used for fragment type.
- */
-void feature_recorder::write_tag(const pos0_t &pos0,size_t len,const string &tagName)
-{
-    stringstream ss;
-    string desc = pos0.alphaPart();
-
-    // This allows you to set a breakpoint at a specific position
-    // it could be a configurable variable, I guess...
-#ifdef DEBUG_OFFSET
-    if(len==DEBUG_OFFSET){
-        std::cerr << "write_tag debug point pos0=" << pos0 << " len=" << len <<" name=" << tagName << "\n";
-    }
-#endif    
-
-    /* offset is either the sbuf offset or the path offset */
-    uint64_t offset = pos0.offset>0 ? pos0.offset : stoi64(pos0.path);
-
-    /** Create what will got to the feature file */
-    ss << offset << ":" << len << "\t";
-    if(desc.size()>0) ss << desc << '/';
-    ss << tagName;
-    
-    this->write(ss.str());
-}
