@@ -9,7 +9,7 @@
 #include <sys/stat.h>
 #include <dirent.h>
 #include <algorithm>
-#include <tr1/unordered_set>
+
 #ifdef HAVE_ERR_H
 #include <err.h>
 #endif
@@ -19,44 +19,21 @@
 #endif
 
 #include "bulk_extractor_i.h"
-
+#include "aftimer.h"
 #include "../dfxml/src/hash_t.h"
 
-namespace std {
-    namespace tr1 {
-        template<>
-        struct hash<md5_t> {
-            size_t operator()(const md5_t &key) const {
-                return *(size_t *)(key.final());
-            }
-        };
-    }
-}
-
-class atomic_hash_set 
-{
-    cppmutex M;
-    std::tr1::unordered_set<md5_t>myset;
-public:
-    atomic_hash_set():M(),myset(){}
-    bool in(const md5_t &s){
-        cppmutex::lock lock(M);
-        return myset.find(s)!=myset.end();
-    }
-    void insert(const md5_t &s){
-        cppmutex::lock lock(M);
-        myset.insert(s);
-    }
-    bool check_for_presence_and_insert(const md5_t &s){
-        cppmutex::lock lock(M);
-        if(myset.find(s)!=myset.end()) return true; // in the set
-        myset.insert(s);                // otherwise insert it
-        return false;                   // and return that it wasn't
-    }
-}seen_md5s;
-#define HAVE_ATOMIC
-
 
+//namespace std {
+//    namespace tr1 {
+//        template<>
+//        struct hash<md5_t> {
+//            size_t operator()(const md5_t &key) const {
+//                return *(size_t *)(key.final());
+//            }
+//        };
+//    }
+//}
+//#endif
 
 uint32_t scanner_def::max_depth = 7;            // max recursion depth
 uint32_t scanner_def::max_ngram = 10;            // max recursion depth
@@ -172,9 +149,7 @@ void be13::plugin::set_scanner_enabled_all(bool enable)
  * The histogram should be done in the plug-in
  */
 
-#ifdef USE_HISTOGRAMS
-static histograms_t histograms;
-#endif
+//static histograms_t histograms;
 
 /****************************************************************
  *** scanner plugin loading
@@ -190,16 +165,18 @@ static histograms_t histograms;
  * This is called before scanners are enabled or disabled, so the pcap handlers
  * need to be set afterwards
  */
-void be13::plugin::load_scanner(scanner_t scanner,const scanner_info::scanner_config &sc)
+void be13::plugin::load_scanner(scanner_t scanner,scanner_info::scanner_config &sc)
 {
     /* If scanner is already loaded, return */
     for(scanner_vector::const_iterator it = current_scanners.begin();it!=current_scanners.end();it++){
         if((*it)->scanner==scanner) return;
     }
 
-    /* make an empty sbuf and feature recorder set */
-    const sbuf_t sbuf;
-    feature_recorder_set fs(feature_recorder_set::SET_DISABLED); // dummy
+    /* Use an empty sbuf and an empty feature recorder set as the parameters for the sp below.
+     * we use static values so that the sbuf is not constantly being created and destroyed.
+     */
+    static const sbuf_t sbuf;
+    static feature_recorder_set fs(feature_recorder_set::SET_DISABLED); // dummy
 
     //
     // Each scanner's params are stored in a scanner_def object that
@@ -226,18 +203,16 @@ void be13::plugin::load_scanner(scanner_t scanner,const scanner_info::scanner_co
     
     sd->enabled      = !(sd->info.flags & scanner_info::SCANNER_DISABLED);
 
-#ifdef USE_HISTOGRAMS
     // Catch histograms and add as a current scanner for all scanners,
     // as a scanner may be enabled after it is loaded.
     for(histograms_t::const_iterator it = sd->info.histogram_defs.begin();
         it != sd->info.histogram_defs.end(); it++){
-        histograms.insert((*it));
+        sc.histograms.insert((*it));
     }
-#endif
     current_scanners.push_back(sd);
 }
 
-void be13::plugin::load_scanner_file(string fn,const scanner_info::scanner_config &sc)
+void be13::plugin::load_scanner_file(string fn,scanner_info::scanner_config &sc)
 {
     /* Figure out the function name */
     size_t extloc = fn.rfind('.');
@@ -277,14 +252,14 @@ void be13::plugin::load_scanner_file(string fn,const scanner_info::scanner_confi
     load_scanner(*scanner,sc);
 }
 
-void be13::plugin::load_scanners(scanner_t * const *scanners, const scanner_info::scanner_config &sc)
+void be13::plugin::load_scanners(scanner_t * const *scanners,scanner_info::scanner_config &sc)
 {
     for(int i=0;scanners[i];i++){
         load_scanner(scanners[i],sc);
     }
 }
 
-void be13::plugin::load_scanner_directory(const string &dirname, const scanner_info::scanner_config &sc )
+void be13::plugin::load_scanner_directory(const string &dirname,scanner_info::scanner_config &sc )
 {
     DIR *dirp = opendir(dirname.c_str());
     if(dirp==0){
@@ -308,7 +283,7 @@ void be13::plugin::load_scanner_directory(const string &dirname, const scanner_i
 }
 
 void be13::plugin::load_scanner_directories(const std::vector<std::string> &dirnames,
-                              const scanner_info::scanner_config &sc)
+                                            scanner_info::scanner_config &sc)
 {
     for(std::vector<std::string>::const_iterator it = dirnames.begin();it!=dirnames.end();it++){
         load_scanner_directory(*it,sc);
@@ -328,25 +303,18 @@ void be13::plugin::load_scanner_packet_handlers()
     }
 }
 
-void be13::plugin::message_enabled_scanners(scanner_params::phase_t phase,feature_recorder_set *fs) // send every enabled scanner the phase message
+// send every enabled scanner the phase message
+void be13::plugin::message_enabled_scanners(scanner_params::phase_t phase,feature_recorder_set &fs) 
 {
-    /* If a feature recorder set wasn't provided, make a dummy */
-    feature_recorder_set *tmpfs=0;
-    if(fs==0){
-        tmpfs = new feature_recorder_set(feature_recorder_set::SET_DISABLED);
-        fs = tmpfs;
-    }
-
     /* make an empty sbuf and feature recorder set */
     const sbuf_t sbuf;
-    scanner_params sp(phase,sbuf,*fs); 
+    scanner_params sp(phase,sbuf,fs); 
     for(scanner_vector::iterator it = current_scanners.begin(); it!=current_scanners.end(); it++){
         if((*it)->enabled){
             recursion_control_block rcb(0,""); // dummy rcb
             ((*it)->scanner)(sp,rcb);
         }
     }
-    if(tmpfs) delete tmpfs;
 }
 
 scanner_t *be13::plugin::find_scanner(const std::string &search_name)
@@ -430,7 +398,7 @@ void be13::plugin::scanners_process_enable_disable_commands()
 }
 
 
-void be13::plugin::scanners_init(feature_recorder_set *fs)
+void be13::plugin::scanners_init(feature_recorder_set &fs)
 {
     message_enabled_scanners(scanner_params::PHASE_INIT,fs); // tell all enabled scanners to init
 }
@@ -439,12 +407,12 @@ void be13::plugin::scanners_init(feature_recorder_set *fs)
  *** PHASE_SHUTDOWN (formerly phase 2): shut down the scanners
  ****************************************************************/
 
-void be13::plugin::phase_shutdown(feature_recorder_set &fs)
+void be13::plugin::phase_shutdown(feature_recorder_set &fs,std::stringstream *sxml)
 {
     for(scanner_vector::iterator it = current_scanners.begin();it!=current_scanners.end();it++){
         if((*it)->enabled){
             const sbuf_t sbuf; // empty sbuf
-            scanner_params sp(scanner_params::PHASE_SHUTDOWN,sbuf,fs);
+            scanner_params sp(scanner_params::PHASE_SHUTDOWN,sbuf,fs,sxml);
             recursion_control_block rcb(0,"");        // empty rcb 
             (*(*it)->scanner)(sp,rcb);
         }
@@ -454,35 +422,12 @@ void be13::plugin::phase_shutdown(feature_recorder_set &fs)
 /****************************************************************
  *** PHASE HISTOGRAM (formerly phase 3): Create the histograms
  ****************************************************************/
-#ifdef USE_HISTOGRAMS
-bool opt_enable_histograms=true;
-void be13::plugin::phase_histogram(feature_recorder_set &fs, xml_notifier_t xml_error_notifier)
-{
-    if(!opt_enable_histograms) return;
-    int ctr = 0;
-    for(histograms_t::const_iterator it = histograms.begin();it!=histograms.end();it++){
-        std::cout << "   " << (*it).feature << " " << (*it).suffix << "...";
-        if(fs.has_name((*it).feature)){
-            feature_recorder *fr = fs.get_name((*it).feature);
-            try {
-                fr->make_histogram((*it));
-            }
-            catch (const std::exception &e) {
-                std::cerr << "ERROR: " ;
-                std::cerr.flush();
-                std::cerr << e.what() << " computing histogram " << (*it).feature << "\n";
-                if(xml_error_notifier){
-                    std::string error = std::string("<error function='phase3' histogram='") + (*it).feature + std::string("</error>");
-                    (*xml_error_notifier)(error);
-                }
-            }
-        }
-        if(++ctr % 3 == 0) std::cout << "\n";
-        std::cout.flush();
-    }
-    if(ctr % 4 !=0) std::cout << "\n";
-}
-#endif
+
+/**
+ * Note currently we have two kinds of histograms:
+ * post-processing histograms specified by the histogram library, and in-memory histograms.
+ * that are really only used by scan_bulk.
+ */
 
 /* option processing */
 /* Get the config and build the help strings at the same time! */
@@ -556,13 +501,14 @@ void scanner_info::get_config(const std::string &n,bool *val,const std::string &
 /**
  * Print a list of scanners.
  * We need to load them to do this, so they are loaded with empty config
+ * Note that scanners can only be loaded once, so this exits.
  */
 void be13::plugin::info_scanners(bool detailed_info,
                                  bool detailed_settings,
                                  scanner_t * const *scanners_builtin,
                                  const char enable_opt,const char disable_opt)
 {
-    const scanner_info::scanner_config empty_config;
+    scanner_info::scanner_config empty_config;
 
     load_scanners(scanners_builtin,empty_config);
     std::cout << "\n";
@@ -659,19 +605,22 @@ void be13::plugin::process_sbuf(const class scanner_params &sp)
 
     /* If we are too deep, error out */
     if(sp.depth >= scanner_def::max_depth){
-        feature_recorder_set::alert_recorder->write(pos0,"process_extract: MAX DEPTH REACHED","");
+        feature_recorder *fr = fs.get_alert_recorder();
+        if(fr) fr->write(pos0,"process_extract: MAX DEPTH REACHED","");
         return;
     }
 
     /* Determine if we have seen this buffer before */
-    md5_t md5 = md5_generator::hash_buf(sp.sbuf.buf,sp.sbuf.bufsize);
-    bool seen_before = seen_md5s.check_for_presence_and_insert(md5);
+    bool seen_before = fs.check_previously_processed(sp.sbuf.buf,sp.sbuf.bufsize);
     if(seen_before){
+        md5_t md5 = md5_generator::hash_buf(sp.sbuf.buf,sp.sbuf.bufsize);
         feature_recorder *alert_recorder = fs.get_alert_recorder();
         std::stringstream ss;
         ss << "<buflen>" << sp.sbuf.bufsize  << "</buflen>";
         if(alert_recorder && dup_data_alerts) alert_recorder->write(sp.sbuf.pos0,"DUP SBUF "+md5.hexdigest(),ss.str());
+#ifdef HAVE__SYNC_ADD_AND_FETCH
         __sync_add_and_fetch(&dup_data_encountered,sp.sbuf.bufsize);
+#endif
     }
 
     /* Determine if the sbuf consists of a repeating ngram. If so,
@@ -717,24 +666,26 @@ void be13::plugin::process_sbuf(const class scanner_params &sp)
                 epath.push_back(toupper(*cc));
             }
 
-            /* Call the scanner.*/
-#ifdef HAVE_AFTIMER
-            aftimer t;
-            t.start();
-#endif
+
             /* Create a RCB that will recursively call process_sbuf() */
             recursion_control_block rcb(process_sbuf,upperstr(name));
-            ((*it)->scanner)(sp,rcb);
-#ifdef HAVE_AFTIMER
-            t.stop();
-            sp.fs.add_stats(epath,t.elapsed_seconds());
-#endif
-                    
-#ifdef DEBUG_PRINT_STEPS
-            if(debug & DEBUG_PRINT_STEPS){
-                cerr << "     ... scanner " << name << " returned\n";
+
+            /* Call the scanner.*/
+            {
+                aftimer t;
+                if(debug & DEBUG_PRINT_STEPS){
+                    cerr << "sbuf.pos0=" << sp.sbuf.pos0 << " calling scanner " << name << "\n";
+                }
+                t.start();
+                ((*it)->scanner)(sp,rcb);
+                t.stop();
+                if(debug & DEBUG_PRINT_STEPS){
+                    cerr << "sbuf.pos0=" << sp.sbuf.pos0 << " scanner "
+                         << name << " t=" << t.elapsed_seconds() << "\n";
+                }
+                sp.fs.add_stats(epath,t.elapsed_seconds());
             }
-#endif
+                
         }
         catch (const std::exception &e ) {
             std::stringstream ss;
@@ -765,7 +716,7 @@ void be13::plugin::process_sbuf(const class scanner_params &sp)
  * Process a pcap packet.
  * Designed to be very efficient because we have so many packets.
  */
-void be13::plugin::process_packet_info(const be13::packet_info &pi)
+void be13::plugin::process_packet(const be13::packet_info &pi)
 {
     for(packet_plugin_info_vector_t::iterator it = packet_handlers.begin(); it != packet_handlers.end(); it++){
         (*(*it).callback)((*it).user,pi);
